import pytest
from datetime import datetime

from aioretry import retry


def retry_policy(fails):
    return False, (fails - 1) * 0.1


@pytest.mark.asyncio
async def test_success():
    @retry(retry_policy)
    async def run(n):
        assert n == 1
        return 1

    assert await run(1) == 1


@pytest.mark.asyncio
async def test_success_instance_normal_rp():
    class A:
        n = 1

        @retry(retry_policy)
        async def run(self):
            assert self.n == 1
            return self.n

    assert await A().run() == 1


@pytest.mark.asyncio
async def test_success_instance_str_rp():
    class A:
        n = 1

        def _retry_policy(self, retries):
            return retry_policy(retries)

        @retry('_retry_policy')
        async def run(self):
            assert self.n == 1
            return self.n

    assert await A().run() == 1


async def run_retry(async_after_failture: bool):
    errors = []

    if async_after_failture:
        async def after_failure(e, fails):
            errors.append(
                (
                    e,
                    fails,
                    datetime.now()
                )
            )

    else:
        def after_failure(e, fails):
            errors.append(
                (
                    e,
                    fails,
                    datetime.now()
                )
            )

    @retry(retry_policy, after_failure)
    async def run():
        length = len(errors)

        if length == 4:
            return 1

        raise RuntimeError(f'{length}')

    current = datetime.now()

    assert await run() == 1

    primative = [
        (int(str(e)), f)
        for e, f, _ in errors
    ]

    assert primative == [
        (0, 1),
        (1, 2),
        (2, 3),
        (3, 4)
    ]

    for i, t in enumerate(errors):
        s = t[2]

        delay = max(0, (i - 1) * 0.1)
        delta = (s - current).total_seconds()

        assert delay < delta < delay + 0.1

        current = s


@pytest.mark.asyncio
async def test_error_normal_after_failure():
    await run_retry(False)


@pytest.mark.asyncio
async def test_error_async_after_failure():
    await run_retry(True)


@pytest.mark.asyncio
async def test_error_usage():
    @retry('_retry_policy')
    async def run():
        return 1

    with pytest.raises(RuntimeError, match=f'retry_policy as a str `"_retry_policy"`'):
        await run()


@pytest.mark.asyncio
async def test_after_failure_fails():
    fail = True

    def after_failure(e, i):
        if fail:
            raise RuntimeError('boom')

    @retry(retry_policy, after_failure)
    async def run():
        if fail:
            raise RuntimeError('haha')

    with pytest.raises(RuntimeError, match='after_failure failed'):
        await run()


@pytest.mark.asyncio
async def test_abandon():
    def retry_policy(fails):
        return fails > 3, fails * 0.1

    fail = True

    @retry(retry_policy)
    async def run():
        if fail:
            raise RuntimeError('boom')

    with pytest.raises(RuntimeError, match='boom'):
        await run()



from typing import (
    Tuple,
    Union,
    Callable,
    Awaitable,
    Optional,
    TypeVar,
    # overload
)

import inspect
import asyncio


RetryPolicyStrategy = Tuple[bool, Union[int, float]]

FunctionRetryPolicy = Callable[[int], RetryPolicyStrategy]
FunctionExceptionCallback = Callable[[Exception, int], Optional[Awaitable]]

RetryPolicy = Union[FunctionRetryPolicy, str]
ExceptionCallback = Union[FunctionExceptionCallback, str]

TargetFunction = Callable[..., Awaitable]

T = TypeVar('T', FunctionRetryPolicy, FunctionExceptionCallback)


async def await_coro(coro):
    if inspect.isawaitable(coro):
        return await coro

    return coro


async def perform(
    fails: int,
    fn: TargetFunction,
    retry_policy: FunctionRetryPolicy,
    after_failure: Optional[FunctionExceptionCallback],
    *args,
    **kwargs
):
    try:
        return await fn(*args, **kwargs)
    except Exception as e:
        fails += 1
        abandon, delay = retry_policy(fails)

        if abandon:
            raise e

        if after_failure is not None:
            try:
                await await_coro(after_failure(e, fails))
            except Exception as e:
                raise RuntimeError(
                    f'[aioretry] after_failure failed, reason: {e}'
                )

        # `delay` could be 0
        if delay:
            await asyncio.sleep(delay)

        return await perform(
            fails,
            fn,
            retry_policy,
            after_failure,
            *args,
            **kwargs
        )


# @overload
# def get_method(
#     target: str,
#     args: Tuple,
#     name: str
# ) -> T:
#     ...

def get_method(
    target: Union[T, str],
    args: Tuple,
    name: str,
) -> T:
    if type(target) is not str:
        return target

    if len(args) == 0:
        raise RuntimeError(
            f'[aioretry] decorator should be used for instance method if {name} as a str `"{target}"` '
        )

    self = args[0]

    return getattr(self, target)  # type: ignore


def retry(
    retry_policy: RetryPolicy,
    after_failure: Optional[ExceptionCallback] = None
) -> Callable[[TargetFunction], TargetFunction]:
    def wrapper(fn: TargetFunction) -> TargetFunction:
        async def wrapped(*args, **kwargs):
            return await perform(
                0,
                fn,
                get_method(
                    retry_policy,
                    args,
                    'retry_policy'
                ),
                get_method(
                    after_failure,
                    args,
                    'after_failure'
                ) if after_failure is not None else None,
                *args,
                **kwargs
            )

        return wrapped

    return wrapper
